#!/usr/bin/python3
"""
    Asterisk AGI Azure OpenAI Application
    This script does the following actions in a loop:
	1. performs streaming speech transcription, 
	2. submits the transcibed text to OpenAI,
	3. synthesizes the text received from OepnAI in response,
	4. plays back the synthesized speech to the caller.
    * Revision: 1
    * Date: Sep 3, 2023
    * Vendor: Universal Speech Solutions LLC
"""

import sys
import requests, json, pycurl
from requests.exceptions import HTTPError
from io import BytesIO
from asterisk.agi import *
from collections import namedtuple
from openai_config import *

class OpenAIApp:

    """A class representing OpenAI application"""

    def __init__(self, options, language):
        """Constructor"""
        self.options_base = options
        self.instance = None
        self.response = None
        self.status = None
        self.cause = None
        self.init_language(language)

    def init_language(self, language):
        """Sets the language"""
        self.language = language
        lang_descr = dict[self.language]
        self.prompt = self.compose_welcome_prompt()
        self.options = self.options_base + '&spl=' + lang_descr.code
        if lang_descr.stt_service_uri:
            self.options += '&vsprec=service-uri=' + lang_descr.stt_service_uri
        if lang_descr.tts_service_uri:
            self.options += '&vspsyn=service-uri=' + lang_descr.tts_service_uri
        agi.verbose('got options %s' % self.options)

    def transcribe_speech(self):
        """Performs a streaming speech transcription"""
        self.grammars = "%s,%s" % (self.compose_speech_grammar(), self.compose_dtmf_grammar())
        self.synth_and_recog()

    def synth_and_recog(self):
        """This is an internal function which calls SynthAndRecog"""
        if not self.prompt:
            self.prompt = ' '
        args = "\\\"%s\\\",\\\"%s\\\",%s" % (
            self.prompt, self.grammars, self.options)
        agi.set_variable('RECOG_STATUS', '')
        agi.set_variable('RECOG_COMPLETION_CAUSE', '')
        self.action = None
        agi.appexec('SynthandRecog', args)
        self.status = agi.get_variable('RECOG_STATUS')
        agi.verbose('got status %s' % self.status)
        if self.status == 'OK':
            self.cause = agi.get_variable('RECOG_COMPLETION_CAUSE')
            agi.verbose('got completion cause %s' % self.cause)
        else:
            agi.verbose('recognition completed abnormally')

    def compose_speech_grammar(self):
        """Composes a built-in speech grammar"""
        grammar = 'builtin:speech/transcribe'
        return grammar

    def compose_dtmf_grammar(self):
        """Composes a built-in DTMF grammar"""
        grammar = 'builtin:dtmf/digits'
        return grammar

    def append_grammar_parameter(self, grammar, name, value, separator):
        """Appends a name/value parameter to the specified grammar"""
        grammar += "%s%s=%s" % (separator, name, value)
        return grammar

    def compose_welcome_prompt(self):
        """Composes a welcome prompt"""
        lang_descr = dict[self.language]
        prompt_collection = lang_descr.prompt_collection
        prompt = "<?xml version=\'1.0\' encoding=\'UTF-8\'?><speak version=\'1.0\' xmlns=\'http://www.w3.org/2001/10/synthesis\' xml:lang=\'%s\'>" % (lang_descr.code)
        prompt += '<p>' + prompt_collection.welcome_prompt  + '</p>'
        prompt += '<p>' + prompt_collection.switch_prompt  + '</p>'
        prompt += '</speak>'
        agi.verbose('got welcome prompt %s' % prompt)
        return prompt

    def compose_next_prompt(self, response):
        """Composes a main prompt"""
        lang_descr = dict[self.language]
        prompt_collection = lang_descr.prompt_collection
        prompt = "<?xml version=\'1.0\' encoding=\'UTF-8\'?><speak version=\'1.0\' xmlns=\'http://www.w3.org/2001/10/synthesis\' xml:lang=\'%s\'>" % (lang_descr.code)
        if response:
            agi.verbose('got json response %s' % response)
            prompt += response['response']
        else:
            prompt += prompt_collection.openai_error_prompt
        prompt += '</speak>'
        agi.verbose('got main prompt %s' % prompt)
        return prompt

    def submit_query(self, text):
        """Submits transcribed text to Open AI"""
        lang_descr = dict[self.language]
        headers = {'Accept':'application/json'}
        payload = "{'query':'%s', 'language':'%s', 'session_id':'%s'}"  % (text, lang_descr.name, agi.env['agi_uniqueid'])
        self.response = None
        agi.verbose('submit request %s to %s' % (payload, lang_descr.openai_service_uri))
        try:
            r = requests.post(lang_descr.openai_service_uri, headers=headers, json=payload, timeout=5)
            agi.verbose('got response code %s' % r.status_code)
            self.response = r.json()
        except HTTPError as http_err:
            agi.verbose('http error occurred: {http_err}')
        except Exception as err:
            agi.verbose('other error occurred: {err}')
        return self.response

    def submit_query_curl(self, text):
        """Submits transcribed text to Open AI using Pycrl"""
        lang_descr = dict[self.language]
        headers = {'Accept':'application/json'}
        payload = "{'query':'%s', 'language':'%s', 'session_id':'%s'}"  % (text, lang_descr.name, agi.env['agi_uniqueid'])
        self.response = None
        agi.verbose('submit request %s to %s' % (payload, lang_descr.openai_service_uri))
        try:
            buffer = BytesIO()
            c = pycurl.Curl()
            c.setopt(c.URL, lang_descr.openai_service_uri)
            c.setopt(pycurl.HTTPHEADER, [ 'Content-Type: application/json' , 'Accept: application/json'])
            data = json.dumps(payload)

            c.setopt(pycurl.POST, 1)
            c.setopt(pycurl.POSTFIELDS, data)
            c.setopt(c.WRITEDATA, buffer)

            c.perform()

            # HTTP response code, e.g. 200.
            agi.verbose('Status: %d' % c.getinfo(c.RESPONSE_CODE))
            # Elapsed time for the transfer.
            agi.verbose('Response time: %f' % c.getinfo(c.TOTAL_TIME))

            c.close()

            self.response = json.loads(buffer.getvalue())
        except Exception as err:
            agi.verbose('Error occurred: {err}')
        return self.response

    def check_dialog_completion(self):
        """Checks whether the dialog is complete"""
        complete = False
        if self.instance == 'Exit' or self.instance == 'Quit' or self.instance == 'exit' or self.instance == 'quit' or self.instance == '0':
            complete = True
        return complete

    def check_language_switch(self):
        """Checks the language switch keyword"""
        lang_descr = dict[self.language]
        switch = False
        if self.instance == lang_descr.switch_keyword:
            switch = True
        return switch

    def run(self):
        """Interacts with the caller in a loop until the dialog is complete"""

        processing = True
        while processing:

            self.transcribe_speech()
            processing = True
            if self.status == 'OK':
                if self.cause == '000':
                    self.instance = agi.get_variable('RECOG_INSTANCE(0/0)')
                    agi.verbose("got instance %s" % self.instance)
                    lang_descr = dict[self.language]
                    if self.check_language_switch():
                        agi.verbose("switching to %s" % lang_descr.switch_to)
                        self.init_language(lang_descr.switch_to)
                        self.prompt = self.compose_welcome_prompt()
                    else:
                        self.response = self.submit_query_curl(self.instance)
                        self.prompt = self.compose_next_prompt(self.response)
                    if self.check_dialog_completion():
                        self.prompt = lang_descr.prompt_collection.final_prompt
                        processing = False

                elif self.cause != '001' and self.cause != '002':
                    processing = False
            elif self.cause != '001' and self.cause != '002':
                processing = False

        agi.appexec('MRCPSynth', "\\\"%s\\\"" % self.prompt)

agi = AGI()
options = 'plt=0&b=1&sct=1500&sint=15000&nit=10000'
language = 'English'
#language = 'Arabic'
openai_app = OpenAIApp(options, language)

openai_app.run()
agi.verbose('exiting')
