#!/usr/bin/python3
"""
    Asterisk AGI Speech Transcription Application
    This script performs speech transcription in a loop by playing the transcribed text back to the caller.
    * Revision: 1
    * Date: Apr 28, 2021
    * Vendor: Universal Speech Solutions LLC
"""

import sys
from asterisk.agi import *
from collections import namedtuple
from config import *

class SpeechTranscriptionApp:

    """A class representing speech transcription application"""

    def __init__(self, options, language):
        """Constructor"""
        self.options_base = options
        self.instance = None
        self.status = None
        self.cause = None
        self.init_language(language)

    def init_language(self, language):
        """Sets the language"""
        self.language = language
        lang_descr = dict[self.language]
        self.prompt = self.compose_welcome_prompt()
        self.options = self.options_base + '&spl=' + lang_descr.code
        if lang_descr.stt_service_uri:
            self.options += '&vsprec=service-uri=' + lang_descr.stt_service_uri
        if lang_descr.tts_service_uri:
            self.options += '&vspsyn=service-uri=' + lang_descr.tts_service_uri
        agi.verbose('got options %s' % self.options)

    def transcribe_speech(self):
        """Performs a streaming speech transcription"""
        self.grammars = "%s,%s" % (self.compose_speech_grammar(), self.compose_dtmf_grammar())
        self.synth_and_recog()

    def synth_and_recog(self):
        """This is an internal function which calls SynthAndRecog"""
        if not self.prompt:
            self.prompt = ' '
        args = "\\\"%s\\\",\\\"%s\\\",%s" % (
            self.prompt, self.grammars, self.options)
        agi.set_variable('RECOG_STATUS', '')
        agi.set_variable('RECOG_COMPLETION_CAUSE', '')
        self.action = None
        agi.appexec('SynthandRecog', args)
        self.status = agi.get_variable('RECOG_STATUS')
        agi.verbose('got status %s' % self.status)
        if self.status == 'OK':
            self.cause = agi.get_variable('RECOG_COMPLETION_CAUSE')
            agi.verbose('got completion cause %s' % self.cause)
        else:
            agi.verbose('recognition completed abnormally')

    def compose_speech_grammar(self):
        """Composes a built-in speech grammar"""
        grammar = 'builtin:speech/transcribe'
        return grammar

    def compose_dtmf_grammar(self):
        """Composes a built-in DTMF grammar"""
        grammar = 'builtin:dtmf/digits'
        return grammar

    def append_grammar_parameter(self, grammar, name, value, separator):
        """Appends a name/value parameter to the specified grammar"""
        grammar += "%s%s=%s" % (separator, name, value)
        return grammar

    def compose_welcome_prompt(self):
        """Composes a welcome prompt"""
        lang_descr = dict[self.language]
        prompt_collection = lang_descr.prompt_collection
        welcome_prompt = prompt_collection.welcome_prompt
        common_prompt = prompt_collection.common_prompt
        prompt = "<?xml version=\'1.0\' encoding=\'UTF-8\'?><speak version=\'1.0\' xmlns=\'http://www.w3.org/2001/10/synthesis\' xml:lang=\'%s\'>" % (lang_descr.code)
        prompt += '<p>' + welcome_prompt  + '</p>'
        prompt += '<p>' + common_prompt  + '</p>'
        prompt += '</speak>'
        agi.verbose('got welcome prompt %s' % prompt)
        return prompt

    def compose_next_prompt(self, text):
        """Composes a main prompt"""
        lang_descr = dict[self.language]
        prompt_collection = lang_descr.prompt_collection
        prompt = "<?xml version=\'1.0\' encoding=\'UTF-8\'?><speak version=\'1.0\' xmlns=\'http://www.w3.org/2001/10/synthesis\' xml:lang=\'%s\'>" % (lang_descr.code)
        prompt += '<p>' + prompt_collection.main_prompt  + "<break time=\'100ms\'/>" + text + '</p>'
        prompt += '<p>' + prompt_collection.common_prompt  + '</p>'
        prompt += '</speak>'
        agi.verbose('got main prompt %s' % prompt)
        return prompt

    def check_dialog_completion(self):
        """Checks whether the dialog is complete"""
        complete = False
        if self.instance == 'Exit' or self.instance == 'Quit' or self.instance == 'exit' or self.instance == 'quit' or self.instance == '0':
            complete = True
        return complete

    def run(self):
        """Interacts with the caller in a loop until the dialog is complete"""

        processing = True
        while processing:

            self.transcribe_speech()
            processing = True
            if self.status == 'OK':
                if self.cause == '000':
                    self.instance = agi.get_variable('RECOG_INSTANCE(0/0)')
                    agi.verbose("got instance %s" % self.instance)
                    lang_descr = dict[self.language]
                    if self.instance == lang_descr.prompt_collection.switch_prompt:
                        agi.verbose("switching to %s" % lang_descr.switch_to)
                        self.init_language(lang_descr.switch_to)
                        self.prompt = self.compose_welcome_prompt()
                    else:
                        self.prompt = self.compose_next_prompt(self.instance)
                    if self.check_dialog_completion():
                        self.prompt = 'Thank you. See you next time.'
                        processing = False

                elif self.cause != '001' and self.cause != '002':
                    processing = False
            elif self.cause != '001' and self.cause != '002':
                processing = False

        agi.appexec('MRCPSynth', "\\\"%s\\\"" % self.prompt)

agi = AGI()
options = 'plt=0&b=1&sct=1500&sint=15000&nit=10000'
language = 'English'
#language = 'Arabic'
transcribe_app = SpeechTranscriptionApp(options, language)

transcribe_app.run()
agi.verbose('exiting')
