#!/usr/bin/python3

import sys
import requests, json, pycurl
from requests.exceptions import HTTPError
from io import BytesIO
from datetime import datetime
from collections import namedtuple
from openai_config import *

class OpenAIApp:

    """A class representing OpenAI application"""

    def __init__(self, options, language):
        """Constructor"""
        self.options_base = options
        self.instance = None
        self.response = None
        self.status = None
        self.cause = None
        self.starttime = None
        self.endtime = None
        self.callerid = '123'
        self.callid = '11-22-33'
        app_counter = '1'
        print(app_counter)
        if app_counter:
            self.app_index = int(app_counter) - 1
        print('new call counter [%d] callerid [%s] callid [%s]' % (self.app_index, self.callerid, self.callid))
        self.init_language(language)

    def init_language(self, language):
        """Sets the language"""
        self.language = language
        self.lang_descr = dict[self.language]
        self.select_stt_service_endpoint()
        self.select_tts_service_endpoint()
        self.select_openai_service_endpoint()
        self.options = self.options_base + '&spl=' + self.lang_descr.code
        if self.stt_service_endpoint:
            self.options += '&vsprec=service-uri=' + self.stt_service_endpoint
        if self.tts_service_endpoint:
            self.options += '&vspsyn=service-uri=' + self.tts_service_endpoint
        print('got options %s' % self.options)

    def select_stt_service_endpoint(self):
        """Selects an STT service endpoint"""
        print('select stt endpoint app_index [%d] len [%d]' % (self.app_index,len(self.lang_descr.service_endpoint_pool_stt)))
        stt_index = self.app_index % len(self.lang_descr.service_endpoint_pool_stt)
        print('got stt index %d' % stt_index)
        self.stt_service_endpoint = self.lang_descr.service_endpoint_pool_stt[stt_index]

    def select_tts_service_endpoint(self):
        """Selects a TTS service endpoint"""
        tts_index = self.app_index % len(self.lang_descr.service_endpoint_pool_tts)
        print('got tts index %d' % tts_index)
        self.tts_service_endpoint = self.lang_descr.service_endpoint_pool_tts[tts_index]

    def select_openai_service_endpoint(self):
        """Selects an OpenAI service endpoint"""
        openai_index = self.app_index % len(self.lang_descr.service_endpoint_pool_openai)
        print('got openai index %d' % openai_index)
        self.openai_service_endpoint = self.lang_descr.service_endpoint_pool_openai[openai_index]

    def start_iteraction(self):
        """Starts a new iteraction with the caller"""
        self.starttime = datetime.now()
        self.transcribe_speech()
        processing = True
        if self.status == 'OK':
            if self.cause == '000':
                # regular processing
                processing = self.process_stt_response('make')
            elif self.cause == '001':
                # no-match
                print("no match")
                self.prompt = self.compose_nomatch_prompt()
            elif self.cause == '002':
                # no-input
                print("no input")
                self.prompt = self.compose_noinput_prompt()
            else:
                # generic error
                self.prompt = self.compose_error_prompt()
                processing = False
        else:
            # generic error
            self.prompt = self.compose_error_prompt()
            processing = False
        self.endtime = datetime.now()
        return processing

    def transcribe_speech(self):
        """Performs a streaming speech transcription"""
        self.grammars = "%s,%s" % (self.compose_speech_grammar(), self.compose_dtmf_grammar())
        self.synth_and_recog()

    def synth_and_recog(self):
        """This is an internal function which calls SynthAndRecog"""
        if not self.prompt:
            self.prompt = ' '
        args = "\\\"%s\\\",\\\"%s\\\",%s" % (
            self.prompt, self.grammars, self.options)

    def process_stt_response(self, instance):
        """This is an internal function which processes speech transcription response"""
        self.instance = instance
        print("got instance %s" % self.instance)
        if self.check_language_switch():
            print("switching to %s" % self.lang_descr.switch_to)
            self.init_language(self.lang_descr.switch_to)
            self.prompt = self.compose_confirmation_prompt()
            return True
        if self.check_dialog_completion():
            self.prompt = self.compose_final_prompt()
            return False

        status = self.submit_query_curl(self.instance)
        if status == False:
            print("got openai error")
            self.prompt = self.compose_openai_error_prompt()
            return False

        self.prompt = self.compose_next_prompt(self.response)
        return True

    def submit_query(self, text):
        """Submits transcribed text to Open AI"""
        headers = {'Accept':'application/json'}
        payload = "{'query':'%s', 'language':'%s', 'session_id':'%s'}"  % (text, self.lang_descr.name, self.callid)
        self.response = None
        status = True
        print('submit request %s to %s' % (payload, self.openai_service_endpoint))
        try:
            r = requests.post(self.openai_service_endpoint, headers=headers, json=payload, timeout=5)
            print('got response code %s' % r.status_code)
            if r.status_code == 200:
                self.response = r.json()
            else:
                status = False
        except HTTPError as http_err:
            print('http error occurred: {http_err}')
            status = False
        except Exception as err:
            print('other error occurred: {err}')
            status = False
        return status

    def submit_query_curl(self, text):
        """Submits transcribed text to Open AI using Pycrl"""
        headers = {'Accept':'application/json'}
        payload = {}
        payload['query'] = text
        payload['language'] = self.lang_descr.name
#        payload['session_id'] = agi.env['agi_uniqueid']
#        payload = "{'query':'%s', 'language':'%s', 'session_id':'%s'}"  % (text, lang_descr.name, self.callid)
        self.response = None
        status = True
        print('submit request %s to %s' % (payload, self.openai_service_endpoint))
        try:
            buffer = BytesIO()
            c = pycurl.Curl()
            c.setopt(c.URL, self.openai_service_endpoint)
            c.setopt(pycurl.HTTPHEADER, [ 'Content-Type: application/json' , 'Accept: application/json'])
            data = json.dumps(payload)

            c.setopt(pycurl.POST, 1)
            c.setopt(pycurl.CONNECTTIMEOUT, 5)
            c.setopt(pycurl.TIMEOUT, 15)
            c.setopt(pycurl.POSTFIELDS, data)
            c.setopt(c.WRITEDATA, buffer)

            c.perform()

            # HTTP response code, e.g. 200.
            print('Status: %d' % c.getinfo(c.RESPONSE_CODE))
            # Elapsed time for the transfer.
            print('Response time: %f' % c.getinfo(c.TOTAL_TIME))
            if c.RESPONSE_CODE != 200:
                status = False

            c.close()

            if status == True:
                self.response = json.loads(buffer.getvalue())
        except pycurl.error as http_err:
            print('http error occurred: %s' % str(http_err))
            status = False
        except Exception as err:
            print('error occurred: %s' % str(err))
            status = False
        return status

    def compose_speech_grammar(self):
        """Composes a built-in speech grammar"""
        grammar = 'builtin:speech/transcribe'
        return grammar

    def compose_dtmf_grammar(self):
        """Composes a built-in DTMF grammar"""
        grammar = 'builtin:dtmf/digits'
        return grammar

    def append_grammar_parameter(self, grammar, name, value, separator):
        """Appends a name/value parameter to the specified grammar"""
        grammar += "%s%s=%s" % (separator, name, value)
        return grammar

    def compose_welcome_prompt(self):
        """Composes a welcome prompt"""
        prompt_collection = self.lang_descr.prompt_collection
        prompt = "<?xml version=\'1.0\' encoding=\'UTF-8\'?><speak version=\'1.0\' xmlns=\'http://www.w3.org/2001/10/synthesis\' xml:lang=\'%s\'>" % (self.lang_descr.code)
        prompt += '<p>' + prompt_collection.welcome_prompt  + '</p>'
        prompt += '<p>' + prompt_collection.switch_prompt  + '</p>'
        prompt += '</speak>'
        print('got welcome prompt %s' % prompt)
        return prompt

    def compose_confirmation_prompt(self):
        """Composes a confirmation prompt of the language change"""
        prompt_collection = self.lang_descr.prompt_collection
        prompt = "<?xml version=\'1.0\' encoding=\'UTF-8\'?><speak version=\'1.0\' xmlns=\'http://www.w3.org/2001/10/synthesis\' xml:lang=\'%s\'>" % (self.lang_descr.code)
        prompt += '<p>' + prompt_collection.confirmation_prompt  + '</p>'
        prompt += '</speak>'
        print('got confirmation prompt %s' % prompt)
        return prompt

    def compose_noinput_prompt(self):
        """Composes a no-input prompt"""
        prompt_collection = self.lang_descr.prompt_collection
        prompt = "<?xml version=\'1.0\' encoding=\'UTF-8\'?><speak version=\'1.0\' xmlns=\'http://www.w3.org/2001/10/synthesis\' xml:lang=\'%s\'>" % (self.lang_descr.code)
        prompt += '<p>' + prompt_collection.noinput_prompt  + '</p>'
        prompt += '</speak>'
        print('got no-input prompt %s' % prompt)
        return prompt

    def compose_nomatch_prompt(self):
        """Composes a no-match prompt"""
        prompt_collection = self.lang_descr.prompt_collection
        prompt = "<?xml version=\'1.0\' encoding=\'UTF-8\'?><speak version=\'1.0\' xmlns=\'http://www.w3.org/2001/10/synthesis\' xml:lang=\'%s\'>" % (self.lang_descr.code)
        prompt += '<p>' + prompt_collection.nomatch_prompt  + '</p>'
        prompt += '</speak>'
        print('got no-match prompt %s' % prompt)
        return prompt

    def compose_error_prompt(self):
        """Composes an error prompt"""
        prompt_collection = self.lang_descr.prompt_collection
        prompt = "<?xml version=\'1.0\' encoding=\'UTF-8\'?><speak version=\'1.0\' xmlns=\'http://www.w3.org/2001/10/synthesis\' xml:lang=\'%s\'>" % (self.lang_descr.code)
        prompt += '<p>' + prompt_collection.error_prompt  + '</p>'
        prompt += '</speak>'
        print('got error prompt %s' % prompt)
        return prompt

    def compose_openai_error_prompt(self):
        """Composes an openai error prompt"""
        prompt_collection = self.lang_descr.prompt_collection
        prompt = "<?xml version=\'1.0\' encoding=\'UTF-8\'?><speak version=\'1.0\' xmlns=\'http://www.w3.org/2001/10/synthesis\' xml:lang=\'%s\'>" % (self.lang_descr.code)
        prompt += '<p>' + prompt_collection.openai_error_prompt  + '</p>'
        prompt += '</speak>'
        print('got openai error prompt %s' % prompt)
        return prompt

    def compose_next_prompt(self, response):
        """Composes a main prompt"""
        prompt_collection = self.lang_descr.prompt_collection
        prompt = "<?xml version=\'1.0\' encoding=\'UTF-8\'?><speak version=\'1.0\' xmlns=\'http://www.w3.org/2001/10/synthesis\' xml:lang=\'%s\'>" % (self.lang_descr.code)
        if response:
            print('got json response %s' % response)
            prompt += '<p>' + response['response']  + '</p>'
            prompt += '<p>' + prompt_collection.switch_prompt  + '</p>'
        else:
            prompt += prompt_collection.openai_error_prompt
        prompt += '</speak>'
        print('got main prompt %s' % prompt)
        return prompt

    def compose_final_prompt(self):
        """Composes a final prompt"""
        prompt_collection = self.lang_descr.prompt_collection
        prompt = "<?xml version=\'1.0\' encoding=\'UTF-8\'?><speak version=\'1.0\' xmlns=\'http://www.w3.org/2001/10/synthesis\' xml:lang=\'%s\'>" % (self.lang_descr.code)
        prompt += '<p>' + prompt_collection.final_prompt  + '</p>'
        prompt += '</speak>'
        print('got final prompt %s' % prompt)
        return prompt

    def check_dialog_completion(self):
        """Checks whether the dialog is complete"""
        complete = False
        if self.instance == 'Exit' or self.instance == 'Quit' or self.instance == 'exit' or self.instance == 'quit' or self.instance == '0':
            complete = True
        return complete

    def check_language_switch(self):
        """Checks the language switch keyword"""
        switch = False
        if self.instance == self.lang_descr.switch_keyword:
            switch = True
        return switch

    def run(self):
        """Interacts with the caller in a loop until the dialog is complete"""
        self.prompt = self.compose_welcome_prompt()

        processing = True
        while processing:
            processing = self.start_iteraction()

#        agi.appexec('MRCPSynth', "\\\"%s\\\"" % self.prompt)


options = 'plt=0&b=1&sct=1500&sint=15000&nit=10000'
language = 'English'
#language = 'Arabic'
openai_app = OpenAIApp(options, language)

openai_app.submit_query_curl('make')
print('exiting')
